import{s as t,b as r,a,S as s}from"./chunk-DI55MBZ5-ItqFkyqz.js";import{_ as i}from"./index-8fiWFYnN.js";import"./chunk-55IACEB6-D09Oskgt.js";import"./chunk-QN33PNHL-DnQMkS7F.js";var l={parser:a,get db(){return new s(2)},renderer:r,styles:t,init:i(e=>{e.state||(e.state={}),e.state.arrowMarkerAbsolute=e.arrowMarkerAbsolute},"init")};export{l as diagram};
